// -*- mode: c++ basic-offset: 2; indent-tabs-mode: nil; -*-
#ifndef DISKCACHE_API__FILENAME_HANDLER_HH
#define DISKCACHE_API__FILENAME_HANDLER_HH

#include "config.h"

#include <list>
#include <map>
#include <memory>
#include <set>

#include "general/types.hh"
#include "general/ReadWriteLock.hh"
#include "general/SharedPtr.hh"

#include "cachefile.hh"
#include "diskcache.hh"

#ifndef ENABLE_MOUNT_POINT_CACHE
#define ENABLE_MOUNT_POINT_CACHE 1
#endif /* ENABLE_MOUNT_POINT_CACHE */

namespace diskCache
{
  //---------------------------------------------------------------------
  // Forward declarations
  //---------------------------------------------------------------------
  class DirectoryInfo;
  class LogMessage;
  class MountPointCache;
  class Query;
  class QueryResults;

  //---------------------------------------------------------------------
  /// \brief Interface for handling different classes of filenames
  //---------------------------------------------------------------------
  class FilenameHandler
  {
  public:
    //-------------------------------------------------------------------
    // Forward declarations
    //-------------------------------------------------------------------
    class Hash;

    typedef General::SharedConstPtr< DirectoryInfo >    dir_info_type;
    typedef General::WeakPtr< DirectoryInfo >           weak_dir_info_type;

    //-------------------------------------------------------------------
    /// \brief Interface class for fast searching of files
    ///
    /// This class provides information that is to be used to organize
    /// a collection of file names within a directory that will be
    /// used to answer queries.
    //-------------------------------------------------------------------
    class Hash
    {
    public:
      //-----------------------------------------------------------------
      /// \brief Constructor
      ///
      /// \param[in] WDirInfo
      ///     Object referencing the directory where this object had
      ///     been originated.
      //-----------------------------------------------------------------
      Hash( weak_dir_info_type WDirInfo );

      //-----------------------------------------------------------------
      /// \brief Destructor
      //-----------------------------------------------------------------
      virtual ~Hash( );

      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      virtual void Read( CacheFile& Stream ) = 0;

      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      virtual void Write( CacheFile& Stream ) const = 0;

      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      virtual void Find( const Query& SearchParams,
                         QueryResults& Results ) const = 0;

      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      virtual size_t FileCount( ) const = 0;

      //-----------------------------------------------------------------
      /// \brief Return emptiness state.
      //-----------------------------------------------------------------
      virtual bool empty( ) const;

    protected:
      weak_dir_info_type     m_directory_info_ref;

    private:
    };

    struct IsFileData
    {
      struct PrivateData
      {
        PrivateData( );

        virtual ~PrivateData( );

        //---------------------------------------------------------------
        // \brief Action to take place once directory has been read.
        //---------------------------------------------------------------
        virtual void OnDirectoryClose( General::SharedPtr< DirectoryInfo > Dir ) = 0;

        INT_4U  s_added_file_count;
        INT_4U  s_removed_file_count;
      };

      typedef General::unordered_map< FilenameHandler*,
                                      General::SharedPtr< PrivateData >,
                                      General::hash_pointer< FilenameHandler > >
      private_data_container_type;


      IsFileData( const DirectoryInfo& PreviousView,
                  General::SharedPtr< DirectoryInfo > CurrentView );

      //-----------------------------------------------------------------
      /// \brief Return the count of files added.
      //-----------------------------------------------------------------
      INT_4U FilesAdded( ) const;

      //-----------------------------------------------------------------
      /// \brief Return the count of files removed.
      //-----------------------------------------------------------------
      INT_4U FilesRemoved( ) const;

      //-----------------------------------------------------------------
      /// \brief Action to take place once directory has been read.
      //-----------------------------------------------------------------
      void OnDirectoryClose( );

      const DirectoryInfo&                      s_previous_view;
      General::SharedPtr< DirectoryInfo >       s_current_view;
      private_data_container_type               s_private_data;
    };

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    typedef  FilenameHandler container_info_key_base_type;
    typedef  container_info_key_base_type* container_info_key_type;
    //-------------------------------------------------------------------
    //
    //-------------------------------------------------------------------
    typedef General::SharedPtr< Hash > hash_value_type;
    typedef std::pair< container_info_key_type, hash_value_type >
    hash_info_type;

    class HashContainer
      : protected std::list< hash_info_type >
    {
    public:
      typedef General::WeakPtr< DirectoryInfo >         weak_dir_info_type;

      typedef std::list< hash_info_type >               container_type;
      typedef container_type::const_iterator            const_iterator;
      typedef container_type::iterator                  iterator;
      typedef container_type::value_type                value_type;
      typedef std::list< std::string >                  dir_container_type;

      using std::list< hash_info_type >::begin;
      using std::list< hash_info_type >::end;
      using std::list< hash_info_type >::push_back;

      //-----------------------------------------------------------------
      /// \brief Constructor
      //-----------------------------------------------------------------
      HashContainer( );

      //-----------------------------------------------------------------
      /// \brief Add all known subdirectories to the global hash.
      //-----------------------------------------------------------------
      void Register( ) const;

      //-----------------------------------------------------------------
      /// \brief Removes all known subdirectories from the global hash.
      ///
      /// \param[out] errors
      ///     Accumulative error message. Appended messsage must be
      ///     enclosed in Tcl " {} " list.
      //-----------------------------------------------------------------
      void Unregister( LogMessage& Errors ) const;

      void Find( const Query& SearchParams,
                 QueryResults& Results ) const;

      //-----------------------------------------------------------------
      /// \brief Retrieve the count of files currently being managed.
      ///
      /// \return
      ///     The number of files being managed.
      //-----------------------------------------------------------------
      size_t FileCount( ) const;

      //-----------------------------------------------------------------
      /// \brief Clear contents
      //-----------------------------------------------------------------
      void clear( );

      //-----------------------------------------------------------------
      /// \brief Read the contents from a CacheFile stream.
      ///
      /// \param[in,out] Stream
      ///     Stream from which to populate the Hash function.
      /// \param[in] WDirInfo
      ///     Directory information from which the object was
      ///     populated.
      //-----------------------------------------------------------------
      void read( CacheFile& Stream,
                 General::SharedPtr< DirectoryInfo > DirInfo );

      //-----------------------------------------------------------------
      /// \brief Wrtie the contents to a CacheFile stream.
      ///
      /// \param[in] Stream
      ///     The Stream to write the contents of the Hash object.
      //-----------------------------------------------------------------
      void write( CacheFile& Stream ) const;

    protected:
      const_iterator find( hash_info_type::first_type Value ) const;
    };

    //-------------------------------------------------------------------
    //-------------------------------------------------------------------
    typedef std::list< FilenameHandler* >       filename_handler_container;

    //-------------------------------------------------------------------
    /// \brief Destructor
    //-------------------------------------------------------------------
    virtual ~FilenameHandler( );

    //-------------------------------------------------------------------
    /// \brief Virtual constructor of Hash object
    ///
    /// Provides a hook for search algorithms to create a search
    /// specific container for files they maintain.
    ///
    /// \param[in] Value
    ///     The directory from which the created hash will be
    ///     associated.
    ///
    /// \return
    ///     The newly created Hash object;
    //-------------------------------------------------------------------
    virtual Hash* CreateHash( dir_info_type Value ) const = 0;

#if ENABLE_MOUNT_POINT_CACHE
    //-------------------------------------------------------------------
    /// \brief Virtual constructor of MountPointCache::Info entry
    //-------------------------------------------------------------------
    virtual MountPointCache* CreateMountPointCache( ) const = 0;
#endif /* ENABLE_MOUNT_POINT_CACHE */

    //-------------------------------------------------------------------
    /// \brief Check if file is parsable by handler
    ///
    /// This method is a hook for the search algorithms to add a
    /// filename to a local collection of searchable files.
    ///
    /// \param[in] Filename
    ///     The file name that is to be considered to be added to a hash.
    /// \param[in,out] Data
    ///     State information.
    ///
    /// \return
    ///     The value true is returned if any
    //-------------------------------------------------------------------
    virtual bool IsFile( const std::string& Filename, IsFileData& Data ) const = 0;
    
    //-------------------------------------------------------------------
    /// \brief Loop over all FilenameHandlers gathering filename list.
    ///
    /// \param[in] Filename
    /// \param[in,out] Data
    ///
    /// \return
    ///     If any registered FilenameHandler returns true from calling
    ///     IsFile, then true is returned; false otherwise.
    //-------------------------------------------------------------------
    static bool IsFileCheck( const std::string& Filename, IsFileData& Data );

    //-------------------------------------------------------------------
    /// \brief Register filename handler with system.
    ///
    /// \param[in] Handler
    ///     FilenameHandler to register.
    ///
    /// \return
    ///     If Handler is added to the FilenameHandler container, the
    ///     value true is returned; false otherwise.
    //-------------------------------------------------------------------
    static bool RegisterHandler( FilenameHandler& Handler );

    //-------------------------------------------------------------------
    /// \brief Register filename handler with system.
    ///
    /// \param[in] Handler
    ///     FilenameHandler to unregister.
    ///
    /// \return
    ///     If Handler is removed from the FilenameHandler container, the
    ///     value true is returned; false otherwise.
    //-------------------------------------------------------------------
    static bool UnregisterHandler( FilenameHandler& Handler );

    //-------------------------------------------------------------------
    /// \brief List of filename handlers
    //-------------------------------------------------------------------
    static filename_handler_container           m_handlers;
    //-------------------------------------------------------------------
    /// \brief Protector of m_handlers.
    ///
    /// This allows access to m_handlers in a thread safe manor.
    //-------------------------------------------------------------------
    static General::ReadWriteLock::Sync         m_handlers_baton;
  };

  //---------------------------------------------------------------------
  /// Construct a new Hash object.
  //---------------------------------------------------------------------
  inline FilenameHandler::Hash::
  Hash( weak_dir_info_type WDirInfo )
    : m_directory_info_ref( WDirInfo )
  {
  }

  //---------------------------------------------------------------------
  /// The hash is considered to be empty if it contains no searchable
  /// files.
  //---------------------------------------------------------------------
  inline bool FilenameHandler::Hash::
  empty( ) const
  {
    return ( FileCount( ) == 0 );
  }


  inline FilenameHandler::HashContainer::const_iterator FilenameHandler::HashContainer::
  find( hash_info_type::first_type Value ) const
  {
    const_iterator cur = begin( ),
      last = end( );
    while( cur != last )
    {
      if ( cur->first == Value )
      {
        break;
      }
      ++cur;
    }
    return cur;
  }

  inline CacheFile& operator<<( CacheFile& Stream,
                                const FilenameHandler::HashContainer& Container )
  {
    Container.write( Stream );
    return Stream;
  }
}
#endif /* DISKCACHE_API__FILENAME_HANDLER_HH */
