#include "config.h"

#include <sstream>
#include <stdexcept>

#include "QueryResultsTimeRange.hh"

namespace /* anonymous */
{
  inline INT_4U ifloor( INT_4U Target, INT_4U Start, INT_4U DT )
  {
    return Start + ( ( ( Target - Start ) / DT ) * DT );
  }

  inline INT_4U iceil( INT_4U Target, INT_4U Stop, INT_4U DT )
  {
    return Stop - ( ( ( Stop - Target ) / DT ) * DT );
  }
} /* namespace - anonymous */

namespace diskCache
{
  //=====================================================================
  //
  //=====================================================================

  QueryResultsTimeRange::RangeData::
  ~RangeData( )
  {
  }

  //=====================================================================
  //
  //=====================================================================

  const Query::gps_time_type QueryResultsTimeRange::m_RESAMPLE_PADDING = 1;

  QueryResultsTimeRange::
  QueryResultsTimeRange( format_type Format, bool AllowGaps )
    : QueryResults( Format, TYPE_TIME_RANGE ),
      m_allow_gaps( AllowGaps )
  {
  }

  QueryResultsTimeRange::
  ~QueryResultsTimeRange( )
  {
  }

  void QueryResultsTimeRange::
  Append( Query::gps_time_type Start,
	  reconstitute_func_type Func,
	  range_data_type	 Data )
  {
    m_file_data[ Start ] = filename_info( Func, Data );
  }

  void QueryResultsTimeRange::
  Finalize( const Query& SearchParams )
  {
    //-------------------------------------------------------------------
    // Start of data set
    //-------------------------------------------------------------------
    switch( format( ) )
    {
    case FORMAT_TCL:
      tcl_list_begin( m_string_result );
      break;
    default:
      // Do nothing
      break;
    }
    //-------------------------------------------------------------------
    // Place each file into result set
    //-------------------------------------------------------------------
    for ( filename_info_container::const_iterator
	    cur = m_file_data.begin( ),
	    last = m_file_data.end( );
	  cur != last;
	  ++cur )
    {
      (*cur->second.s_func)(cur->first, cur->second.s_data.get( ), *this );
    }
    //-------------------------------------------------------------------
    // End of data set
    //-------------------------------------------------------------------
    switch( format( ) )
    {
    case FORMAT_TCL:
      tcl_list_end( m_string_result );
      break;
    default:
      // Do nothing
      break;
    }
    //-------------------------------------------------------------------
    // Put in conflict data
    //-------------------------------------------------------------------
    tcl_list_begin( m_string_result );
    if ( m_conflicts.empty( ) == false )
    {
      m_string_result == " ";
      m_string_result += m_conflicts;
    }
    if ( ( m_allow_gaps == false )
	 && ( SearchParams.TimeRange( ).empty( ) == false ) )
    {
      //-----------------------------------------------------------------
      // Report the gappy information
      //-----------------------------------------------------------------
      tcl_list_begin( m_string_result );
      m_string_result += "Gaps are detected:";
      for ( Query::time_range_type::const_iterator
	      cur_gap = SearchParams.TimeRange( ).begin( ),
	      last_gap = SearchParams.TimeRange( ).end( );
	    cur_gap != last_gap;
	    ++cur_gap )
      {
	std::ostringstream	gap;
	gap << " " << cur_gap->first << " " << cur_gap->second;

	m_string_result += gap.str( );
      }
      tcl_list_end( m_string_result );
    }
    tcl_list_end( m_string_result );
  }

  void QueryResultsTimeRange::
  Finalize( const Query& SearchParams, TimeHash::KnownIntervals& NewData )
  {
    typedef TimeHash::KnownIntervals::ranges_type::mapped_type data_info_type;

    switch( format( ) )
    {
    case FORMAT_CONFLICT:
      //-----------------------------------------------------------------
      // Each of these represents conflicting data
      //-----------------------------------------------------------------
      for ( filename_info_container::const_iterator
	      cur = m_file_data.begin( ),
	      last = m_file_data.end( );
	    cur != last;
	    ++cur )
      {
	Query::gps_time_type	cs( cur->first );
	Query::gps_time_type	ce( cur->second.s_data->EndTime( ) );
	INT_4U			cdt( cur->second.s_data->DeltaT( ) );
        TimeHash::KnownIntervals::ranges_type::iterator
	  del;

        for ( TimeHash::KnownIntervals::ranges_type::iterator
                nd_cur = NewData.Ranges( ).begin( ),
                nd_last = NewData.Ranges( ).end( );
              nd_cur != nd_last;
              )
        {
	  Query::gps_time_type	ds( nd_cur->first );
	  Query::gps_time_type	de( nd_cur->second.mStopTime );
	  INT_4U		ddt( nd_cur->second.mFileDt );

	  if ( ce < ds )
	  {
	    //-----------------------------------------------------------
	    // No conflict can happen because the file ends before
	    // the remaining data sets start.
	    //-----------------------------------------------------------
	    break;
	  }
	  if ( cs > de )
	  {
	    //-----------------------------------------------------------
	    // Conflicting data starts after the known interval so
	    // advance to the next known interval.
	    //-----------------------------------------------------------
	    ++nd_cur;
	    continue;
	  }
	  //-------------------------------------------------------------
	  // Remove conflicts from list of known intervals.
	  // \todo Generate error messages describing conflicts.
	  //-------------------------------------------------------------
	  INT_4U	crs( 0 );	// conflict resolution start
	  INT_4U	cre( 0 );	// conflict resolution end

	  INT_4U	cds( 0 );	// conflicting data start
	  INT_4U	cde( 0 );	// conflicting data end

	  if ( ds >= cs )
	  {
	    //-----------------------------------------------------------
	    // C2 or C3
	    //-----------------------------------------------------------
	    crs = ifloor( ds, cs, cdt );
	    cds = ds;
	    del = nd_cur;
	    if ( de > ce )
	    {
	      //---------------------------------------------------------
	      // C3
	      //---------------------------------------------------------
	      // # Conflict (new): [830100015,830100045:30)
	      // #     with (old): [830100000,830100030:30)
	      // touch conflicting_data/1_old_data/Z-C3_A-830100000-30.gwf
	      // touch conflicting_data/2_new_data/Z-C3_A-830100015-30.gwf
	      // @ # Conflict (new): [830100015,830100030:15)
	      // @ #     with (old): [830100000,830100030:30)
	      // @ touch conflicting_data/1_old_data/Z-C3_A1-830101000-30.gwf
	      // @ touch conflicting_data/2_new_data/Z-C3_A1-830101015-15.gwf
	      // @ touch conflicting_data/2_new_data/Z-C3_A1-830101030-15.gwf
	      // # Conflict (new): [830200016,830200032:16)
	      // #     with (old): [830200000,830200048:48)
	      // touch conflicting_data/1_old_data/Z-C3_B-830200000-15.gwf
	      // touch conflicting_data/2_new_data/Z-C3_B-830200000-30.gwf
	      cre = ce;
	      cde = iceil( de, ce, cdt );

	      INT_4U s( iceil( ce, de, ddt ) );

	      if ( s < de )
	      {
		TimeHash::KnownIntervals::ranges_type::value_type
		  n( s,
		     data_info_type( de, ddt ) );
		nd_cur = NewData.Ranges( ).insert( n ).first;
	      }
	    }
	    else
	    {
	      //---------------------------------------------------------
	      // de <= ce
	      // C2
	      //---------------------------------------------------------
	      // # Conflict (new): [820100016,820100032:16)
	      // #     with (old): [820100000,820100048:48)
	      // touch conflicting_data/1_old_data/Z-C2_A-820100000-48.gwf
	      // touch conflicting_data/2_new_data/Z-C2_A-820100016-16.gwf
	      // # Conflict (new): [820101016,820101032:16)
	      // #     with (old): [820101000,820101048:24)
	      // touch conflicting_data/1_old_data/Z-C2_A1-820101000-24.gwf
	      // touch conflicting_data/1_old_data/Z-C2_A1-820101024-24.gwf
	      // touch conflicting_data/2_new_data/Z-C2_A1-820101016-16.gwf
	      // # Conflict (new): [820200016,820200048:32)
	      // #     with (old): [820200000,820200048:48)
	      // touch conflicting_data/1_old_data/Z-C2_B-820200000-48.gwf
	      // touch conflicting_data/2_new_data/Z-C2_B-820200016-32.gwf
	      // # Conflict (new): [820300000,820300032:32)
	      // #     with (old): [820300000,820300048:48)
	      // touch conflicting_data/1_old_data/Z-C2_C-820300000-48.gwf
	      // touch conflicting_data/2_new_data/Z-C2_C-820300000-32.gwf
	      // # Conflict (new): [820400000,820400048:48)
	      // #     with (old): [820400000,820400048:48)
	      // touch conflicting_data/1_old_data/Z-C2_D-820400000-48.gwf
	      // touch conflicting_data/2_new_data/Z-C2_D-820400000-48.gwf
	      cre = iceil( de, ce, cdt );
	      cde = de;
	    }
	    ++nd_cur;
	    NewData.Ranges( ).erase( del );
	    goto conflict_error_check;
	  }
	  else
	  {
	    //-----------------------------------------------------------
	    // ds < cs
	    // C1 or C4
	    //-----------------------------------------------------------
	    if ( de > cs )
	    {
	      crs = cs;
	      cds = ifloor( cs, ds, ddt );

	      nd_cur->second.mStopTime = ifloor( cs, ds, ddt );
	      if ( de > ce )
	      {
		//-------------------------------------------------------
		// C1
		//-------------------------------------------------------
		// Conflict (new): [810000000,810000048:48)
		//     with (old): [810000016,810000032:16)
		// touch conflicting_data/1_old_data/Z-C1_A-810000016-16.gwf
		// touch conflicting_data/2_new_data/Z-C1_A-810000000-48.gwf
		// # Conflict (new): [810000016,810000032:16)
		// #     with (old): [810000016,810000032:16)
		// touch conflicting_data/1_old_data/Z-C1_A-810000016-16.gwf
		// touch conflicting_data/2_new_data/Z-C1_A-810000000-16.gwf
		// touch conflicting_data/2_new_data/Z-C1_A-810000016-16.gwf
		// touch conflicting_data/2_new_data/Z-C1_A-810000032-16.gwf
		INT_4U s( iceil( ce, de, ddt ) );

		cre = ce;
		cde = s;

		TimeHash::KnownIntervals::ranges_type::value_type
		  n( s,
		     data_info_type( de, ddt ) );
		nd_cur = NewData.Ranges( ).insert( n ).first;
	      }
	      else
	      {
		//-------------------------------------------------------
		// de <= ce
		// C4
		//-------------------------------------------------------
		// # Conflict (new): [840100000,840100032:32)
		// #     with (old): [840100016,840100048:32)
		// touch conflicting_data/1_old_data/Z-C4_A-840100016-32.gwf
		// touch conflicting_data/2_new_data/Z-C4_A-840100000-32.gwf
		// # Conflict (new): [840101014,840101028:14)
		// #     with (old): [840101016,840101048:32)
		// touch conflicting_data/1_old_data/Z-C4_A1-840101016-32.gwf
		// touch conflicting_data/2_new_data/Z-C4_A1-840101000-14.gwf
		// touch conflicting_data/2_new_data/Z-C4_A1-840101014-14.gwf
		// # Conflict (new): [840200000,840200032:32)
		// #     with (old): [840200016,840200032:16)
		// touch conflicting_data/1_old_data/Z-C4_B-840200016-16.gwf
		// touch conflicting_data/2_new_data/Z-C4_B-840200000-32.gwf
		// # Conflict (new): [840202010,840202030:10)
		// #     with (old): [840202015,840202030:15)
		// touch conflicting_data/1_old_data/Z-C4_B1-840201015-15.gwf
		// touch conflicting_data/2_new_data/Z-C4_B1-840201000-10.gwf
		// touch conflicting_data/2_new_data/Z-C4_B1-840201010-10.gwf
		// touch conflicting_data/2_new_data/Z-C4_B1-840201020-10.gwf
		cre = ce;
		cde = de;
		if ( ds == nd_cur->second.mStopTime )
		{
		  del = nd_cur;
		  ++nd_cur;
		  NewData.Ranges( ).erase( del );
		  goto conflict_error_check;
		}
	      }
	    }
	  }
	  ++nd_cur;
	conflict_error_check:
	  //-------------------------------------------------------------
	  // Check if a conflict has been detected.
	  //-------------------------------------------------------------
	  if ( cde != 0 )
	  {
	    //-----------------------------------------------------------
	    // Some conflicts exists so need to generate the log
	    // message.
	    //-----------------------------------------------------------
	    std::cerr << "DEBUG: Conflicting: "
		      << " data: [" << cds << ", " << cde << " {" << ddt << "})"
		      << " conflicts with: [" << crs << ", " << cre << " {" << cdt << "})"
		      << std::endl
	      ;
	  }
	}
      }
      break;
    default:
      break;
    }
  }

  bool QueryResultsTimeRange::
  Resample( Query& SearchParams ) const
  {
    if ( (m_file_data.empty( ) ) // No results
	 || ( SearchParams.TimeRange( ).empty( ) == false ) ) // gappy
    {
      //-----------------------------------------------------------------
      // Cannot resample a query that has gaps
      //-----------------------------------------------------------------
      return false;
    }

    //-------------------------------------------------------------------
    // Expand the search by N seconds on each side of the query.
    //-------------------------------------------------------------------
    Query::time_range_type	retval;
    filename_info_container::const_iterator
      head = m_file_data.begin( ),
      tail = m_file_data.end( );

    --tail;

    retval[ head->first - m_RESAMPLE_PADDING ] = head->first;
    retval[ tail->second.s_data->EndTime( ) ] = tail->second.s_data->EndTime( ) + m_RESAMPLE_PADDING;

    SearchParams.TimeRange( retval );

    return true;
  }

} // namespace - diskCache
