#ifndef DiskCacheCmdHH
#define DiskCacheCmdHH   

// #define DEBUG_TIME

// System Header Files   
#include <string>
#include <iostream>
   
#include "general/types.hh"
#include "genericAPI/threaddecl.hh"

#if FOR_TCL
#include "genericAPI/tid.hh"
#endif /* FOR_TCL */

//------------------------------------------------------------------------------
//
//: Invoke "slow" scan (recursive collection of subdirectories and 
//: frame data files) on directory listed in MOUNT_PT list (rsc file variable).
//   
// This function gets recursive content of the passed to it directory. It
// should be called only for directories that appear in the MOUNT_PT list.   
// The same function is used to introduce new MOUNT_PT directory or to rescan
// already existing directory in the global hash for new data.
//   
//!usage: set string [ getDirEnt dirpath ]
//   
//!param: const char* dirpath - A full path to the directory.
//   
//!return: const string - A list of all identified and scanned subdirectories
//+        that have frame data.   
//      
std::string getDirEnt( const std::string dirpath );

#if defined( CREATE_THREADED1_DECL )
CREATE_THREADED1_DECL( getDirEnt, std::string, const std::string );   
#endif /* defined( CREATE_THREADED1_DECL ) */

   
//------------------------------------------------------------------------------
//   
//: Update list of mount point directories (as they appear in
//: the resource file): initialize list if called at API startup time;
//:                     otherwise, check if any MOUNT_PT directory
//:                     has been pruned, remove that directory from
//:                     global hash.   
//
//!usage: set dir_list [ updateMountPtList known_mount_pt_dirs optional_global_check ]
//   
//!param: const CHAR* dir_list - A list of directories as they appear
//+       in MOUNT_PT list (API resource file variable).
//!param: const bool enable_global_check - A flag to indicate if API should 
//+       check for data conflicts across all MOUNT_PT entries 
//+       (diskcacheAPI@LDAS reports data conflicts only under the same 
//+       MOUNT_PT).   
//+       This flag should be set to TRUE only when called by outside of 
//+       LDAS utility, such as /ldas/bin/cacheCheck.   
//+       Default is FALSE.      
//   
//!return: string - Sorted Tcl list of all removed subdirectories, followed
//+        by error messages if any: {dir1 dir2 ...} {error1 error2 ...}    
//      
std::string updateMountPtList( const char* dir_list, 
                               const bool enable_global_check = false );

//------------------------------------------------------------------------------
//   
//: Get ASCII representation of the current frame hash. 
//
//!param: const char* ifo - An ifo to look up. Default is "all".
//!param: const char* type - A type to look up. Default is "all".
//   
//!return: string - A list of all identified subdirectories that contain frame 
//+        data with a Tcl-layer requested metadata, such as:
//+        DirectoryName,IFO,Type,NumberOfFramesPerFile(always 1),dT(file), 
//+        MTIME NUMBER_OF_FRAMES_IN_DIR {StartTime StopTime ...}  
//   
std::string getDirCache( const char* ifo = "all", const char* type = "all" );   
#if ! FOR_PYTHON
CREATE_THREADED2_DECL( getDirCache, std::string, const char*, const char* );
#endif /* ! FOR_PYTHON */


#if ! FOR_PYTHON
//------------------------------------------------------------------------------
//  
//: Get TCL formatted lists for each mount point with name or mount point, 
//: number of directories and number of files for data matching the specified
//: ifo and type.  Return lists are of the form:
//:   {mountpoint_name number_of_dirs number_of_files } 
//
//!usage: set string [ getHashNumbers ]
//
//!param: const char* ifo - An ifo to look up. Default is "all".
//!param: const char* type - A type to look up. Default is "all".
//
//!return: string - A list for each mount point with name, number of
//+        directories and number of files under that mount point:
//+        {mountpoint_name number_of_dirs number_of files }
// 
std::string getHashNumbers( const char* ifo = "all", const char* type = "all" );
#endif /* ! FOR_PYTHON */


//------------------------------------------------------------------------------
//
//: Get TCL formatted lists of file names for data matching the specified
//: IFO and Type within the bounds of the query.
//
//!usage: set string [ getFileNames ifo_type_str query_start query_stop ]
//
//!param: const char* ifo_type_str - A space delimited list of IFO-Type strings
//!param: const unsigned int start_time - query start time.
//!param: const unsigned int stop_time - query stop time.
//
//!return: string - A list for each mount point with data file names:
//+ { mntpt1 { /f1 /f2 ... /fN } mntpt2 { /f1 /f2 ... /fN } ... mntptN { } }  
//
std::string getFileNames(
           const char* ifo_type_str,
           const unsigned int query_start,
           const unsigned int query_stop);

//------------------------------------------------------------------------------
//
//: Get TCL formatted lists of intervals for data matching the specified 
//: IFO and Type within the bounds of the query.  
//
//!usage: set string [ getIntervalsList ifo_type_str query_start query_stop ]
//
//!param: const char* ifo_type_str - A space delimited list of IFO-Type strings.
//!param: const unsigned int start_time - query start time.
//!param: const unsigned int stop_time - query stop time.   
//
//!return: string - A list for each IFO-Type pair with data intervals:
//+  IFO-Type1 { i1_start i1_stop ... iN_start iN_stop } ... IFO-TypeN { ... }
//
std::string getIntervalsList( 
                const char* ifo_type_str, 
                const unsigned int query_start,
                const unsigned int query_stop);

#if FOR_LDAS
CREATE_THREADED3_DECL( getIntervalsList, std::string, const char*, 
		       const unsigned int, const unsigned int );
#endif /* ! FOR_PYTHON && ! FOR_TCL */


std::string getFrameFiles( const char* ifo, const char* type,
                           const unsigned int start_time, 
                           const unsigned int stop_time,
                           const bool gaps_allowed );       
   
#if FOR_LDAS
CREATE_THREADED5_DECL( getFrameFiles, std::string, const char*, const char*,
                       const unsigned int, const unsigned int, const bool );
#endif /* ! FOR_PYTHON && ! FOR_TCL */

   
//------------------------------------------------------------------------------
//   
//: RDS frame query.
//   
// This command finds frame files that satisfy user request. If "will_resample"
// flag is set to TRUE, requested time range will be automatically expanded by
// frame dt at the beginning and at the end of the range.
//   
//!param: const char* ifo - Data IFO.    
//!param: const char* type - Data Type.       
//!param: const unsigned int start_time - Data start time.
//!param: const unsigned int stop_time - Data stop time.   
//!param: const bool will_resample - Flag to indicate if requested data will be
//+                                  resampled. Set to TRUE (1) if will be 
//+                                  resampled, to FALSE (0) if not.
//!return: string - Tcl formated string: { frame files }{ errors if any }.
//                    
std::string getRDSFrameFiles( const char* ifo, const char* type,
                              const unsigned int start_time, 
                              const unsigned int stop_time,
                              const bool will_resample );       
   
#if FOR_LDAS
CREATE_THREADED5_DECL( getRDSFrameFiles, std::string, const char*, const char*,
                       const unsigned int, const unsigned int, const bool );   
#endif /* ! FOR_PYTHON && ! FOR_TCL */
   

//------------------------------------------------------------------------------
//   
//: Write content of global frame data hash to binary file.   
//
//!param: const char* filename - Name of the file to write frame hash to.
//+       Default is an empty string (C++ will use default file name).      
//
//!return: Nothing.
//
void writeDirCache( const char* filename = "" );   
#if ! FOR_PYTHON
CREATE_THREADED1V_DECL( writeDirCache, const char* );         
#endif /* ! FOR_PYTHON */

//------------------------------------------------------------------------------
//   
//: Write content of global frame data hash to binary file.   
//
//!param: const char* filename - Name of the file to write frame hash to.
//+       Default is an empty string (C++ will use default file name).      
//
//!return: Nothing.
//
void writeDirCacheAscii( const char* filename = "" );   
#if ! FOR_PYTHON
CREATE_THREADED1V_DECL( writeDirCacheAscii, const char* );         
#endif /* ! FOR_PYTHON */

   
//------------------------------------------------------------------------------
//   
//: Write content of global frame data hash to binary file.   
//
//!param: const char* bfilename - Name of the file to write binary frame hash to.
//+       Default is an empty string (C++ will use default file name).      
//
//!param: const char* afilename - Name of the file to write ascii frame hash to.
//+       Default is an empty string (C++ will use default file name).      
//
//!return: Nothing.
//
void writeDirCacheFiles( const char* bfilename = "",
			 const char* afilename = "" );
#if ! FOR_PYTHON
CREATE_THREADED2V_DECL( writeDirCacheFiles, const char*, const char* );
#endif /* ! FOR_PYTHON */

//------------------------------------------------------------------------------
//   
//: Read content of global frame data hash from binary file.   
//
// Tcl layer can specify different files for read and write operations. 
// This function forces all read and write operations to be sequencial.
//   
// ATTN: This function destroys existing hash before reading a new one from the
//       given file. Caller must assure there are no running threads that might
//       access global frame data hash at the time "readDirCache" is called.
//    
//!param: const char* filename - Name of the file to read frame hash from.
//+       Default is an empty string (C++ will use default file name).      
//
//!return: Nothing.
//
void readDirCache( const char* filename = "" );   

#if ! FOR_PYTHON
CREATE_THREADED1V_DECL( readDirCache, const char* );         
#endif /* ! FOR_PYTHON */
   

//------------------------------------------------------------------------------
//   
//: Update list of excluded directories (as they appear in
//: the resource file): check if existing data hash relies on such directories
//:                     already, remove those directories recursively from
//:                     global hash.   
//
//!usage: set dir_list [ excludedDirList dirs_to_exclude ]
//   
//!param: const CHAR* dir_list - A list of directories to exclude 
//+                              (as they appear in API resource file variable).
//   
//!return: string - Sorted Tcl list of all removed subdirectories, followed
//+        by error messages if any:
//+        {Directories excluded by pattern 'dir_list': dir1 dir2 ...} 
//+        {error1 error2 ...}
//         
std::string excludedDirList( const char* dir_list );   


//------------------------------------------------------------------------------
//  
//: Update list of file extensions
//
//!usage: updateFileExtList [ list ]
//  
//!param: const CHAR* ext_list - A list of file extensions 
//  
//!return: nothing 
//
void updateFileExtList( const char* ext_list );


//----------------------------------------------------------------------------
//
//: Obtain the list of file name extensions
//
//!usage: set ext_list [ getFileExtList ]
//
//!param: none
//
//!return: std::string
//
std::string getFileExtList();



   
//------------------------------------------------------------------------------
//   
//: Delete global frame data hash.
//
// ATTN: This function destroys existing hash in memory. 
//       Caller must assure there are no running threads that might
//       access global frame data hash at the time "deleteDirCache" is called.
//    
//!return: Nothing.
//
void deleteDirCache();         
   
#if ! FOR_PYTHON
CREATE_THREADED0V_DECL( deleteDirCache );
#endif /* !FOR_PYTHON */

#if FOR_TCL
bool bindToCheckForMountPTConflictVariable( Tcl_Interp* Interp, char* Variable );
#endif /* FOR_TCL */


//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
void DumpCacheDaemonStart( std::ostream* Stream = NULL );

//------------------------------------------------------------------------------
/// \brief The number of mount point directories to scan concurrently
//------------------------------------------------------------------------------
void ScanConcurrency( INT_4U Concurrency );

//------------------------------------------------------------------------------
/// \brief Scan the list of mount points.
//------------------------------------------------------------------------------
void ScanMountPointList( std::ostream* Stream = NULL );

//-----------------------------------------------------------------------
/// \brief Continuously scan the list of mount points
//-----------------------------------------------------------------------
void ScanMountPointListContinuously( std::ostream* Stream = NULL );

#endif /* define DiskCacheCmdHH */
