/* -*- c-basic-offset: 3; -*- */

#include <cassert>

// System Header Files
#include <stdexcept>   
#include <sstream>

#include "genericAPI/Logging.hh"

// Local Header Files
#include "diskcacheAPI.hh"
#include "timehash.hh"
#include "cachefile.hh"   
#include "filename.hh"   
#include "mountpt.hh"   
#include "Query.hh"
   
using namespace diskCache;
using namespace General;   
using namespace std;
using GenericAPI::queueLogEntry;
using GenericAPI::LogEntryGroup_type;

   
// Useful typedefs   
// vector< pair >::const_iterator
typedef map< INT_4U, list< INT_4U > >::const_iterator MapConstIter;   

int DISKCACHE_API_EXTRA_VALIDATION = 20;
   
//-----------------------------------------------------------------------
/// \brief Constructor   
///
/// This is the default constructor.
//-----------------------------------------------------------------------
diskCache::TimeHash::KnownIntervals::
KnownIntervals( )   
{
}

//-----------------------------------------------------------------------
/// \brief Constructor
///
/// FIX PR 2390: new query method: pass start,stop into search
///
/// \param[in] ranges A set of known ranges.
//-----------------------------------------------------------------------
diskCache::TimeHash::KnownIntervals::
KnownIntervals( const std::map< INT_4U, IntervalInfo > ranges )
   : mRanges ( ranges )
{
}

//-----------------------------------------------------------------------
/// \brief Copy time ranges from known hash.
///
/// This method is a friend of TimeHash class.
///   
/// \param[in] ifo_hash A pointer to the object representing
///       known time ranges for a particular directory for a particular
///       "IFO-Type-".
///   
/// \return Nothing.
//-----------------------------------------------------------------------
void diskCache::TimeHash::KnownIntervals::
setIntervals( const TimeHash* const ifo_hash )   
{
   if( ifo_hash == 0 )
   {
      return;
   }

   
   mRanges = ifo_hash->getIntervals();
   return;
}


//-----------------------------------------------------------------------
/// \brief Set time ranges from Start Time and IntervalInfo.
///
/// This method is a friend of TimeHash class.
/// FIX PR 2390: new query method: pass start,stop into search
///
/// \param[in] start start time 
/// \param[in] ranges interval info
///
/// \return Nothing.
//-----------------------------------------------------------------------
void diskCache::TimeHash::KnownIntervals::
setIntervals( INT_4U start, IntervalInfo ranges )
{
   mRanges[start] = ranges;
   return;
}

INT_4U diskCache::TimeHash::KnownIntervals::
CountFiles( ) const
{
   INT_4U	retval = 0;

   for ( ranges_type::const_iterator
	    cur_range = mRanges.begin( ),
	    last_range = mRanges.end( );
	 cur_range != last_range;
	 ++cur_range )
   {
      //-----------------------------------------------------------------
      // Calculate the number of files by the equation:
      // ( stop_time - start_time ) / dt
      //-----------------------------------------------------------------
      retval += ( ( cur_range->second.mStopTime - cur_range->first )
		  / cur_range->second.mFileDt );
   }

   return retval;
}



//-----------------------------------------------------------------------
/// \brief Default constructor   
///
/// FIX PR 2390: new query method: pass start,stop into search
//-----------------------------------------------------------------------
diskCache::TimeHash::IntervalInfo::
IntervalInfo()
   : mStopTime( 0 ),
     mFileDt( -1 )
{
}
   
   
//-----------------------------------------------------------------------
/// \brief Constructor   
///
/// \param[in] stop_time Stop time for the whole range of available 
///       frame data.
/// \param[in] dt Interval covered by frame file, in seconds.   
//-----------------------------------------------------------------------
diskCache::TimeHash::IntervalInfo::
IntervalInfo( const INT_4U stop_time, const INT_4U dt )   
   : mStopTime( stop_time ),
     mFileDt( dt )
{
}
   
//-----------------------------------------------------------------------
/// \brief Write object to the stream.
///
/// \param[in] stream A reference to the output stream.
///   
/// \return Nothing.
//-----------------------------------------------------------------------
void diskCache::TimeHash::IntervalInfo::
write( CacheFile& stream ) const
{
   stream << mStopTime << mFileDt;
   
   return;
}
   
//-----------------------------------------------------------------------
/// \brief Read object from the stream.
///
/// \param[in] stream A reference to the input stream.
///   
/// \return Nothing.
//-----------------------------------------------------------------------
void diskCache::TimeHash::IntervalInfo::
read( CacheFile& stream )
{
   stream >> mStopTime >> mFileDt;
   
   return;
}
   
//-----------------------------------------------------------------------
/// \brief Constructor   
//-----------------------------------------------------------------------
diskCache::TimeHash::
TimeHash()
{
   ReadWriteLock::Initialize( mLock );
}
   
//-----------------------------------------------------------------------
/// \brief Destructor   
//-----------------------------------------------------------------------
diskCache::TimeHash::
~TimeHash()
{
   ReadWriteLock::Destroy( mLock );
}
   
size_t diskCache::TimeHash::
Count( ) const
{
   size_t	retval( 0 );

   for( MConstIter
	   iter = mIntervals.begin( ),
	   end_iter = mIntervals.end( );
        iter != end_iter;
	++iter )
   {
      // number of objects = ( stop - start ) / dt 
      retval += ( ( iter->second.mStopTime - iter->first ) / iter->second.mFileDt );
   }
   return retval;
}

//-----------------------------------------------------------------------
/// Adds a set of known intervals
//-----------------------------------------------------------------------
void TimeHash::
Add( const KnownIntervals& Source )
{
   for ( KnownIntervals::ranges_type::const_iterator
	    s_cur = Source.Ranges( ).begin( ),
	    s_last = Source.Ranges( ).end( );
	 s_cur != s_last;
	 ++s_cur )
   {
      interval_container_type::reverse_iterator
	 l_cur = mIntervals.rbegin( ),
	 l_pre = mIntervals.rend( ),
	 l_last = mIntervals.rend( );
      while( l_cur != l_last )
      {
	 if ( l_cur->second.mFileDt == s_cur->second.mFileDt )
	 {
	    if ( s_cur->first > l_cur->second.mStopTime )
	    {
	       if ( ( l_pre != l_last )
		    && ( s_cur->second.mStopTime == l_pre->first ) )
	       {
		  interval_container_type::value_type
		     v( s_cur->first,
			interval_container_type::mapped_type( l_pre->second.mStopTime,
							      s_cur->second.mFileDt ) );
		  mIntervals.erase( (++l_pre).base( ) );
		  mIntervals.insert( v );
	       }
	       else
	       {
		  //-----------------------------------------------------
		  // Add new node since the data is not part of the set
		  //-----------------------------------------------------
		  mIntervals.insert( *s_cur );
	       }
	       break;
	    }
	    else if ( s_cur->first == l_cur->second.mStopTime )
	    {
	       if ( ( l_pre != l_last )
		    && ( l_pre->first == s_cur->second.mStopTime ) )
	       {
		  //-----------------------------------------------------
		  // Filling in a gap
		  //-----------------------------------------------------
		  l_cur->second.mStopTime = l_pre->second.mStopTime;
		  mIntervals.erase( (++l_pre).base( ) );
	       }
	       else
	       {
		  //-----------------------------------------------------
		  // Modify the node since the data extends the set
		  //-----------------------------------------------------
		  l_cur->second.mStopTime = s_cur->second.mStopTime;
	       }
	       break;
	    }
	    l_pre = l_cur;
	 }
	 ++l_cur;
      }
      if ( l_cur == l_last )
      {
	 if ( ( l_pre != l_last )
	      && ( l_pre->first == s_cur->second.mStopTime ) )
	 {
	    //-----------------------------------------------------------
	    // This is extending at the front end a segment.
	    //-----------------------------------------------------------
	    interval_container_type::value_type	v( *s_cur );
	    v.second.mStopTime = l_pre->second.mStopTime;
	    mIntervals.erase( (++l_pre).base( ) );
	    mIntervals.insert( v );
	 }
	 else
	 {
	    //-----------------------------------------------------------
	    // Simply add new node at the beginning of the list.
	    //-----------------------------------------------------------
	    mIntervals.insert( *s_cur );
	 }
      }
   }
}

//-----------------------------------------------------------------------
/// Adds an entry from a single file.
//-----------------------------------------------------------------------
void TimeHash::
Add( INT_4U GPSStart, INT_4U DT )
{
#if 1
   KnownIntervals	k;

   k.setIntervals( GPSStart, IntervalInfo( GPSStart + DT, DT ) );

   Add( k );
#else
   ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
		       __FILE__, __LINE__ );

   const register INT_4U gpsend( GPSStart + DT );

   interval_container_type::reverse_iterator
      l_cur = mIntervals.rbegin( ),
      l_pre = mIntervals.rend( ),
      l_last = mIntervals.rend( );
   while( l_cur != l_last )
   {
      if ( l_cur->second.mFileDt == DT )
      {
	 if ( GPSStart > l_cur->second.mStopTime )
	 {
	    if ( ( l_pre != l_last )
		 && ( gpsend == l_pre->first ) )
	    {
	       interval_container_type::value_type
		  v( GPSStart,
		     interval_container_type::mapped_type( l_pre->second.mStopTime, DT ) );
	       mIntervals.erase( (++l_pre).base( ) );
	       mIntervals.insert( v );
	    }
	    else
	    {
	       //--------------------------------------------------------
	       // Add new node since the data is not part of the set
	       //--------------------------------------------------------
	       interval_container_type::value_type
		  v( GPSStart,
		     interval_container_type::mapped_type( gpsend, DT ) );
	    
	       mIntervals.insert( v );
	    }
	    break;
	 }
	 else if ( GPSStart == l_cur->second.mStopTime )
	 {
	    if ( ( l_pre != l_last )
		 && ( l_pre->first == gpsend ) )
	    {
	       //--------------------------------------------------------
	       // Filling in a gap
	       //--------------------------------------------------------
	       l_cur->second.mStopTime = l_pre->second.mStopTime;
	       mIntervals.erase( (++l_pre).base( ) );
	    }
	    else
	    {
	       //--------------------------------------------------------
	       // Modify the node since the data extends the set
	       //--------------------------------------------------------
	       l_cur->second.mStopTime = gpsend;
	    }
	    break;
	 }
	 l_pre = l_cur;
      }
      ++l_cur;
   }
   if ( l_cur == l_last )
   {
      //-----------------------------------------------------------------
      // Assume the simple case of the data set describing a new node.
      //-----------------------------------------------------------------
      interval_container_type::value_type
	 v( GPSStart,
	    interval_container_type::mapped_type( gpsend,
						  DT ) );
      if ( ( l_pre != l_last )
	   && ( l_pre->first == gpsend ) )
      {
	 //--------------------------------------------------------------
	 // This is extending at the front end a segment.
	 //--------------------------------------------------------------
	 v.second.mStopTime = l_pre->second.mStopTime;
	 //--------------------------------------------------------------
	 // Remove the node that has been absorbed into the new node.
	 //--------------------------------------------------------------
	 mIntervals.erase( (++l_pre).base( ) );
      }
      //-----------------------------------------------------------------
      // Add the new node into the container
      //-----------------------------------------------------------------
      mIntervals.insert( v );
   }
#endif /* 0 */
}

//-----------------------------------------------------------------------
/// This method returns true if an entry specified by GPSStart time
/// and DT exists; false otherwise.
//-----------------------------------------------------------------------
bool diskCache::TimeHash::
Find( INT_4U GPSStart, INT_4U DT ) const
{
   const register INT_4U endtime( GPSStart + DT );
   bool retval( false );

   for( interval_container_type::const_iterator
	   cur = mIntervals.begin( ),
	   last = mIntervals.end( );
	( cur != last )
	   && ( cur->first <= GPSStart );
	++cur )
   {
      if ( ( cur->second.mFileDt )
	   && ( cur->second.mStopTime >= endtime ) )
      {
	 retval = true;
	 break;
      }
   }
   return retval;
}

void diskCache::TimeHash::
Find( const Query& SearchParams,
      KnownIntervals& Intervals ) const
{

   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
		       __FILE__, __LINE__ );

   const Query::bit_flag_type criteria( SearchParams.Criteria( ) );

   if ( criteria & Query::CRITERIA_BIT_TIME_RANGE )
   {
      if ( mIntervals.empty( )
	   || SearchParams.TimeRange( ).empty( ) )
      {
	 return;
      }
	  
      Query	sp( SearchParams );

      Query::time_range_type& tr( sp.TimeRange( ) );

      Query::time_range_type::iterator
	 tr_cur = tr.begin( ),
	 tr_last = tr.end( );
      --tr_last;

      //------------------------------------------------------------------
      // Find the bounds of the local time intervals.
      //------------------------------------------------------------------
      MConstIter
	 // capture the lower bound
	 i_cur = mIntervals.lower_bound( tr_cur->first ),
	 // captgure the upper bound
	 i_last = mIntervals.upper_bound( tr_last->second );
      if ( ( i_cur == mIntervals.end( ) )
	   || ( ( i_cur->first != tr_cur->first )
		&& ( i_cur != mIntervals.begin( ) ) ) )
      {
	 --i_cur;
      }
      ++tr_last;

      while ( ( i_cur != i_last )
	      && ( tr_cur != tr_last ) )
      {
	 //--------------------------------------------------------------
	 // Skip over nodes that do not need to be checked because
	 // they are completely out of range.
	 //--------------------------------------------------------------
	 while ( ( tr_cur->second <= i_cur->first )
		 || ( tr_cur->first >= i_cur->second.mStopTime ) )
	 {
	    while ( tr_cur->second <= i_cur->first )
	    {
	       //--------------------------------------------------------
	       // Skip elements that are out of range
	       //--------------------------------------------------------
	       ++tr_cur;
	       if ( tr_cur == tr_last )
	       {
		  //-----------------------------------------------------
		  // Nothing more to find
		  //-----------------------------------------------------
		  return;
	       }
	    }
	    while ( tr_cur->first >= i_cur->second.mStopTime )
	    {
	       //--------------------------------------------------------
	       // The start of what is being looked for happens after
	       //   this time interval.
	       //--------------------------------------------------------
	       ++i_cur;
	       if ( i_cur == i_last )
	       {
		  //-----------------------------------------------------
		  // Nothing more to find
		  //-----------------------------------------------------
		  return;
	       }
	    }
	 }
	 //--------------------------------------------------------------
	 // Add Elements to list of known intervals
	 //--------------------------------------------------------------
	 if ( ( tr_cur->first >= i_cur->first )
	      && (tr_cur->second <= i_cur->second.mStopTime ) )
	 {
	    //-----------------------------------------------------------
	    // The entirety of the time range request can be satisfied
	    //   by this entry
	    //-----------------------------------------------------------
	    Query::gps_time_type start
	       = tr_cur->first - ( ( tr_cur->first - i_cur->first )
				   % i_cur->second.mFileDt );
	    Query::gps_time_type stop
	       = tr_cur->second + ( ( i_cur->second.mStopTime
				      - tr_cur->second )
				    % i_cur->second.mFileDt );
	    Intervals.mRanges[ start ]
	       = IntervalInfo( stop, i_cur->second.mFileDt );
	    Query::time_range_type::iterator	r = tr_cur;
	    ++tr_cur;
	    tr.erase( r );
	    continue;
	    
	 }
	 else if ( ( tr_cur->first >= i_cur->first )
		   && ( tr_cur->second > i_cur->second.mStopTime ) )
	 {
	    //-----------------------------------------------------------
	    // The lower portion of the query can be satisfied by
	    // this interval
	    //-----------------------------------------------------------
	    Query::gps_time_type start
	       = tr_cur->first - ( ( tr_cur->first - i_cur->first )
				   % i_cur->second.mFileDt );

	    Intervals.mRanges[ start ]
	       = IntervalInfo( i_cur->second.mStopTime,
			       i_cur->second.mFileDt );

	    Query::time_range_type::iterator	r = tr_cur;
	    //-----------------------------------------------------------
	    // Create new node of data
	    //-----------------------------------------------------------
	    tr_cur = tr.insert( tr_cur, 
				Query::time_range_type::
				value_type( i_cur->second.mStopTime,
					    tr_cur->second ) );
	    //-----------------------------------------------------------
	    // Remove previous node of data
	    //-----------------------------------------------------------
	    tr.erase( r );
	    ++i_cur;
	    continue;
	 }
	 else if ( ( tr_cur->first < i_cur->first )
		   && ( tr_cur->second <= i_cur->second.mStopTime ) )
	 {
	    //-----------------------------------------------------------
	    // The upper portion of the query can be satisfied by
	    // this interval
	    //-----------------------------------------------------------
	    Query::gps_time_type stop
	       = tr_cur->second + ( ( i_cur->second.mStopTime
				      - tr_cur->second )
				    % i_cur->second.mFileDt );
	    Intervals.mRanges[ i_cur->first ]
	       = IntervalInfo( stop, i_cur->second.mFileDt );
	    tr_cur->second = i_cur->first;	// Reset end
	    ++tr_cur;
	    continue;
	 }
	 else if ( ( tr_cur->first < i_cur->first )
		   && ( tr_cur->second > i_cur->second.mStopTime ) )
	 {
	    Intervals.mRanges[ i_cur->first ]
	       = IntervalInfo( i_cur->second.mStopTime, i_cur->second.mFileDt );
	    //-----------------------------------------------------------
	    // Split the search time range into two pieces.
	    //-----------------------------------------------------------
	    Query::gps_time_type end( tr_cur->second );
	    tr_cur->second = i_cur->first;
	    tr_cur = tr.insert( tr_cur,
				Query::time_range_type::
				value_type( i_cur->second.mStopTime,
					    end ) );
	    ++i_cur;
	    continue;
	    
	 }
	 //--------------------------------------------------------------
	 // Advance to the next interval of cached data
	 //--------------------------------------------------------------
	 ++i_cur;
      }
   }
   else if ( ( criteria
	       & ( Query::CRITERIA_BIT_START_TIME | Query::CRITERIA_BIT_END_TIME ) )
	     == 0 )
   {
      //-----------------------------------------------------------------
      // The answer is everything
      //-----------------------------------------------------------------
      Intervals.mRanges = mIntervals;
   }
   else
   {
	
      // Loop over all data intervals for this timehash object
      INT_4U query_start( ( Query::CRITERIA_BIT_START_TIME & criteria )
			  ? SearchParams.StartTime( )
			  : 0 );
      INT_4U query_stop( ( Query::CRITERIA_BIT_END_TIME & criteria )
			 ? SearchParams.EndTime( )
			 : ~0 );

      for( MConstIter
	      iter = mIntervals.begin( ),
	      end_iter = mIntervals.end( );
	   iter != end_iter;
	   ++iter )
      {
	 // iter->first            is the start time of the data interval
	 // iter->second.mStopTime is the stop  time of the data interval

	 // Only count data intervals that are partially in, wholy in, 
	 // or exceed the quary's start and stop times (q1, q2):
	 //
	 //   q1   q2       q1    q2 
	 //    |   |         |    | 
	 //  - - - - -      --------
	 //  a b c d e         f     
	 //
	 // case a: data interval completely outside the bounds of the query
	 // case b: data interval straddles the start time of the query
	 // case c: data interval inside the bounds of the query
	 // case d: data interval straddles the stop time of the query
	 // case e: data interval completely outside the bounds of the query
	 // case f: data interval larger than bounds of the query 

	 // cases a and e should NOT be added to the collection maps

	 // all other cases are partially (b,d) , wholly (c), or exceed (f)
	 // the bounds of the query and should be added
 
	 if ( iter->second.mStopTime < query_start ) 
	 {
	    // case a: data interval outside the bounds of the query (before)
	    // do not add data to collection maps
	    continue; // to next interval for this timehash object
	 }
	 else if ( iter->first > query_stop )
	 {
	    // case e: data interval outside the bounds of the query (after)
	    // do not add data to helper maps
	    continue; // to next interval for this timehash object
	 }
	 else 
	 {
	    const INT_4U	dt( iter->second.mFileDt );

	    //-----------------------------------------------------------
	    // The start is either the start of the interval
	    // OR it is the query_start rounded down to the nearest
	    // interval start.
	    //-----------------------------------------------------------
	    INT_4U	start( ( iter->first > query_start )
			       ? iter->first
			       : query_start - ( query_start % dt ) );
	    //-----------------------------------------------------------
	    // The stop is either the stop of the interval
	    // OR it is the query_stop rounded up to the nearest
	    // interval stop.
	    //-----------------------------------------------------------
	    INT_4U	end( ( iter->second.mStopTime < query_stop )
			     ? iter->second.mStopTime
			     : query_stop + ( dt - ( query_stop % dt ) ) );
	    Intervals.mRanges[ start ] = IntervalInfo( end, dt );
	 }

      } // end of loop over mIntervals
   }
}

//-----------------------------------------------------------------------
/// \brief Is data hash empty?
///
/// \return True if data hash is empty, false otherwise.
//-----------------------------------------------------------------------
const bool diskCache::TimeHash::
empty() const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
		       __FILE__, __LINE__ );      
   
   return mIntervals.empty();
}
   
   
//-----------------------------------------------------------------------
/// \brief Get Tcl-layer requested representation of known frame data.
///
/// \param[in] record
///       "DirectoryName,IFO,Type,NumberOfFramesPerFile,"
///       part of ascii record.
///       NumberOfFramesPerFile currently is always 1.
/// \param[in] mtime Mtime for the directory data resides under.   
///
/// \return const std::string - Complete frame record.
//-----------------------------------------------------------------------
const std::string diskCache::TimeHash::
getTclHash( const std::string& record,
	    const INT_4S mtime ) const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
		       __FILE__, __LINE__ );      

   // Since data map is based on start time of data range, scan through 
   // all entries and sort them by FileDt as requested by Tcl-layer
   // dT <---> vector( iterators into mIntervals )
   map< INT_4U, vector< MConstIter > > dt_map;
   
   // Tcl-layer requested info: dT <---> number of frame files for that dT
   map< INT_4U, INT_4U > dt_numfiles_map;
   
   
   for( MConstIter iter = mIntervals.begin(), end_iter = mIntervals.end();
        iter != end_iter; ++iter )
   {
      if ( iter->second.mFileDt <= 0 )
      {
	 //--------------------------------------------------------------
	 // :ERROR: This should never be the case
	 //--------------------------------------------------------------
	 std::cerr << "WARNING: range error for mFileDt:"
		   << " record: '" << record << "'"
		   << " mtime: " << mtime
		   << " start Time: " << iter->first
		   << " mStopTime: " << iter->second.mStopTime
		   << " mFileDt: " << iter->second.mFileDt
		   << " mIntervals.size(): " << mIntervals.size( )
		   << std::endl;
	 INT_4U j = 0;
	 for( MConstIter
		 i = mIntervals.begin(),
		 e = mIntervals.end();
	      i != e; ++i, ++j )
	 {
	    std::cerr << "   mIntervals[" << j << "]:"
		      << " start Time: " << iter->first
		      << " mStopTime: " << iter->second.mStopTime
		      << " mFileDt: " << iter->second.mFileDt
		      << std::endl;
	 }
	 continue;
      }
      dt_map[ iter->second.mFileDt ].push_back( iter );
     
      // Calculate how many files are in the range:
      //     num_files = ( StopTime - StartTime ) / dT
      if( dt_numfiles_map.find( iter->second.mFileDt ) == dt_numfiles_map.end() )
      {
         // Initialize entry: 
         dt_numfiles_map[ iter->second.mFileDt ] = 
              ( iter->second.mStopTime - iter->first ) / iter->second.mFileDt;
      }
      else
      {
         // Update entry
         dt_numfiles_map[ iter->second.mFileDt ] += 
             ( iter->second.mStopTime - iter->first ) / iter->second.mFileDt;
      }
   }
   
   
   ostringstream data;
   
   // Usefull typedef
   typedef map< INT_4U, vector< MConstIter > >::iterator dtIterator;
   
   for( dtIterator iter = dt_map.begin(), end_iter = dt_map.end();
        iter != end_iter; ++iter )
   {
      // Format "dt MTIME NUMBER_OF_FRAMES_IN_DIR {"
      data << record << iter->first << ' '
           << mtime << ' ' 
           << dt_numfiles_map[ iter->first ] << " {";
      
   
      const vector< MConstIter >& intervals( iter->second );
   
   
      // Format time ranges
      if( intervals.empty() == false )
      {
         vector< MConstIter >::const_iterator interval_iter = intervals.begin();
      
         data << ( *interval_iter )->first             // interval StartTime
              << ' '
              << ( *interval_iter )->second.mStopTime; // interval StopTime
      
         ++interval_iter;
      
      
         for( vector< MConstIter >::const_iterator interval_end_iter = intervals.end(); 
              interval_iter != interval_end_iter; ++interval_iter )
         {
            data << ' '
                 << ( *interval_iter )->first             // interval StartTime
                 << ' '
                 << ( *interval_iter )->second.mStopTime; // interval StopTime
         }
      }
   
   
      data << "} ";
   }   
   
   
   return data.str();
}

//-----------------------------------------------------------------------
/// \brief Write object to the stream.
///
/// \param[in] stream A reference to the output stream.
///   
/// \return Nothing.
//-----------------------------------------------------------------------
void diskCache::TimeHash::
write( CacheFile& stream ) const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
		       __FILE__, __LINE__ );         

   
   const INT_4U hash_size( mIntervals.size() );   
   
   // Write hash size
   stream << hash_size;

   
   for( MConstIter iter = mIntervals.begin(), end_iter = mIntervals.end();
        iter != end_iter; ++iter )
   {
      // Write start_time
      stream << iter->first;
   
      // Write IntervalInfo
      iter->second.write( stream );
   }
   
   
   return;
}

   
//-----------------------------------------------------------------------
/// \brief Read object from the stream.
///
/// \param[in] stream A reference to the input stream.
///   
/// \return Nothing.
//-----------------------------------------------------------------------
void diskCache::TimeHash::
read( CacheFile& stream )
{
   static const char caller[] = "diskCache::TimeHash::read";

   // Read hash size
   INT_4U hash_size;
   
   stream >> hash_size;

   if ( DISKCACHE__CACHE_FILE_VERBOSE_DEBUGGING )
   {
      std::ostringstream	msg;

      msg << "hash_size: " << hash_size;

      queueLogEntry( msg.str( ),
		     LogEntryGroup_type::MT_DEBUG,
		     30,
		     caller,
		     "CXX_INTERFACE" );
   }
   
   if( hash_size )
   {
      ReadWriteLock lock( mLock, ReadWriteLock::WRITE,
                          __FILE__, __LINE__ );            
   
      while( hash_size )
      {
         // Read start_time
         INT_4U key;
         stream >> key;
   
	 IntervalInfo	i;
	 i.read( stream );
	 if ( i.mFileDt > 0 )
	 {
	    // Insert new entry for the key 
	    mIntervals.insert( MValueType( key, i ) );
	 }
   
   
         // Read one map entry
         --hash_size;
      }
   }
   
   
   return;
}

   
//-----------------------------------------------------------------------
/// \brief Get a copy of mIntervals.
///
/// \return A copy of mIntervals.
//-----------------------------------------------------------------------
const std::map< INT_4U, TimeHash::IntervalInfo > diskCache::TimeHash::getIntervals() 
   const
{
   ReadWriteLock lock( mLock, ReadWriteLock::READ, RWLOCK_TIMEOUT,
		       __FILE__, __LINE__ );         
   
   return mIntervals;
}

//-----------------------------------------------------------------------
/// \brief Add element to time hash
///
/// \param[in] Pos Position to put Value.
/// \param[in] Value The data to be inserted.
///
/// \return The iterator just past the inserted dat.a
//-----------------------------------------------------------------------
TimeHash::MIter TimeHash::
insert( TimeHash::MIter Pos, const TimeHash::MValueType& Value )
{
   if ( DISKCACHE_API_EXTRA_VALIDATION )
   {
      //-----------------------------------------------------------------
      // Check for duplicate
      //-----------------------------------------------------------------
      if ( ( Pos->first == Value.first )
	   && ( Pos->second.mStopTime == Value.second.mStopTime )
	   && ( Pos->second.mFileDt == Value.second.mFileDt ) )
      {
	 std::cerr << "WARNING: Duplicate interval entry:"
		   << std::endl
		   << "\tstart(pos): " << Pos->first
		   << " mStopTime: " << Pos->second.mStopTime
		   << " mFileDt: " << Pos->second.mFileDt
#if DISKCACHE_API__TIME_HASH_DEBUG
		   << " now: " << Pos->second.m_when
		   << " thread_id: " << Pos->second.m_thread_id
		   << " ( now: " << General::GPSTime::NowGPSTime( )
		   << " thread_id: " << pthread_self( )
		   << " )"
#endif /* DISKCACHE_API__TIME_HASH_DEBUG */
		   << std::endl;
      }
      if ( DISKCACHE_API_EXTRA_VALIDATION > 10 )
      {
	 assert ( ! ( ( Pos->first == Value.first )
		      && ( Pos->second.mStopTime == Value.second.mStopTime )
		      && ( Pos->second.mFileDt == Value.second.mFileDt ) ) );

	 assert( Value.second.mFileDt > 0 );
	 for ( MConstIter
		  cur = mIntervals.begin( ),
		  last = mIntervals.begin( );
	       cur != last;
	       ++cur )
	 {
	    assert( cur->second.mFileDt > 0 );
	    assert( ! ( ( cur->first == Value.first )
			&& ( cur->second.mFileDt == Value.second.mFileDt )
			&& ( cur->second.mStopTime == Value.second.mStopTime ) ) );
	 }
      }
   }

   Pos =  mIntervals.insert( Pos, Value );
   if ( DISKCACHE_API_EXTRA_VALIDATION )
   {
#if DISKCACHE_API__TIME_HASH_DEBUG
      //-----------------------------------------------------------------
      // Update debugging information
      //-----------------------------------------------------------------
      Pos->second.m_thread_id = pthread_self( );
      Pos->second.m_when.Now( );
#endif /* DISKCACHE_API__TIME_HASH_DEBUG */
   }
   return Pos;
}

//-----------------------------------------------------------------------
/// This method returns true if an entry specified by GPSStart time
/// and DT exists; false otherwise.
//-----------------------------------------------------------------------
bool diskCache::TimeHash::
Overlapping( INT_4U GPSStart, INT_4U GPSEnd ) const
{
   bool retval( false );

   for( interval_container_type::const_reverse_iterator
	   cur = mIntervals.rbegin( ),
	   last = mIntervals.rend( );
	( cur != last )
	   && ( cur->first >= GPSEnd );
	++cur )
   {
      if ( ( ( GPSStart >= cur->first )
	     && ( GPSStart < cur->second.mStopTime ) )
	   || ( ( GPSEnd > cur->first )
		&& ( GPSEnd <= cur->second.mStopTime ) ) )
      {
	 retval = true;
	 break;
      }
   }
   return retval;
}
