#if HAVE_CONFIG_H
#include "framecpp/config.h"
#endif /* HAVE_CONFIG_H */

#include <iomanip>
#include <memory>
#include <sstream>

#include "general/AtExit.hh"
#include "general/unittest.h"

#include "framecpp/Common/FrameH.hh"

#include "framecpp/Version8/FrameH.hh"

#include "framecpp/FrameCPP.hh"
#include "framecpp/OFrameStream.hh"

using std::filebuf;
using std::ofstream;

using namespace FrameCPP;

typedef Common::FrameBuffer< filebuf > FrameBuffer;
typedef Common::FrameH frameh_type;

static const INT_4U CHANNEL_COUNT = 64;
static const General::GPSTime START_TIME( 913423140, 0 );
static const INT_2U ULEAPS = START_TIME.GetLeapSeconds( );
static const REAL_8 DT = 60.0;

General::UnitTest 	Test;

static std::auto_ptr< frameh_type> build_frame( version_type Version,
						INT_4U ChannelCount );
int
main( int argc, char* argv[] )
{
  Test.Init( argc, argv );

  try
  {
    int frame_spec = FRAME_SPEC_CURRENT;
    int channel_count = CHANNEL_COUNT;

    //-------------------------------------------------------------------
    // Construct a frame
    //-------------------------------------------------------------------
    std::auto_ptr< frameh_type >
      frameh( build_frame( frame_spec, channel_count ) );
    //-------------------------------------------------------------------
    // Constuct output frame name
    //-------------------------------------------------------------------
    std::ostringstream	filename;
    filename << "Z-SixtyK_" << channel_count
	     << "-" << START_TIME.GetSeconds( )
	     << "-" << int( DT )
	     << ".gwf";
    //-------------------------------------------------------------------
    // Time writing of the frame
    //-------------------------------------------------------------------
    General::GPSTime	start;
    start.Now( );

    ofstream	ffile( filename.str( ).c_str( ) );

    std::auto_ptr< FrameBuffer >
      obuf( new FrameBuffer( std::ios::out ) );
    obuf->open( filename.str( ).c_str( ),
		std::ios::out | std::ios::binary );
    OFrameStream	ofs( obuf.release( ) );

    ofs.WriteFrame( *frameh );
    ofs.Close( );

    General::GPSTime	finish;
    finish.Now( );
    Test.Check( ( finish - start ) < DT )
      << "Writing of frame completed in less than real time"
      << std::endl
      ;
    Test.Message( 10 )
      << "Time to write file: " << ( finish - start )
      << " Percent of Real Time: " << ( ( finish - start ) * 100 / DT )
      << std::endl
      ;
  }
  catch ( const std::exception& E )
  {
    Test.Check( false ) << "Caught an exception: " << E.what( ) << std::endl;
  }
  catch ( ... )
  {
    Test.Check( false ) << "Caught an unknown exception" << std::endl;
  }
  try
  {
    General::AtExit::Cleanup( );
  }
  catch( ... )
  {
  }
  Test.Exit();
}

//=======================================================================
// Local functions
//=======================================================================

static void add_channels( version_type Version,
			  std::auto_ptr< frameh_type >& Frame,
			  INT_2U padding,
			  INT_4U ChannelCount );

static std::auto_ptr< frameh_type > create_frameh( version_type Version );

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
template< typename RawData_type > void
add_channel( RawData_type& RawData, INT_2U Padding, INT_4U ChannelOffset );

template<>
void
add_channel
( FrameCPP::Version_8::FrRawData& RawData,
  INT_2U Padding,
  INT_4U ChannelOffset )
{
  using namespace FrameCPP::Version_8;

  std::ostringstream	name;
  name << "Z1:Channel"
       << std::setw( Padding ) << std::setfill( '0' ) << ChannelOffset
    ;
  std::vector< REAL_4 > d( size_t( DT ), 0.0 );

  Dimension	dim( d.size( ) );
  FrVect	v( name.str( ), 1,
		   &dim, &d[0] );
  FrAdcData	a( name.str( ),	/* name */
		   0, ChannelOffset, /* group, channel */
		   32,		/* nBits */
		   1,		/* sampleRate */
		   0,		/* bias */
		   1,		/* slope */
		   "NONE",	/* units */
		   0,		/* fShift */
		   0,		/* timeOffset */
		   0,		/* dataValid */
		   0		/* phase */
		   );
  // a.RefData( ).append( &v, false /* allocate */, false /* own */ );
  a.RefData( ).append( v );
  RawData.RefFirstAdc( ).append( a );
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------
template< typename RawData_type >
void add_adc_channels( RawData_type& RawData, INT_2U Padding, INT_4U ChannelCount )
{
  for ( INT_4U cc = 0;
	cc < ChannelCount;
	++cc )
  {
    add_channel( RawData, Padding, cc );
  }
  
}
//-----------------------------------------------------------------------
// Creation of the header
//-----------------------------------------------------------------------
template< int > std::auto_ptr< frameh_type > frameh( );

template<> std::auto_ptr< frameh_type >
frameh< 8 >( )
{
  using namespace FrameCPP::Version_8;

  std::auto_ptr< frameh_type >
    retval;

  std::auto_ptr< FrameH >
    f ( new FrameH( "SixtyK",
		    0,
		    0,
		    START_TIME,
		    ULEAPS,
		    DT ) );
  std::auto_ptr< FrRawData >	r( new FrRawData( "SixtyK" ) );
  f->SetRawData( r.release( ),
		 false /* allocate */,
		 true /* owns */ );
  retval.reset( f.release( ) );
  return retval;
}

//-----------------------------------------------------------------------
//-----------------------------------------------------------------------

static void
add_channels( version_type Version,
	      std::auto_ptr< frameh_type >& Frame,
	      INT_2U Padding,
	      INT_4U ChannelCount )
{
  switch( Version )
  {
  case 8:
    {
      using namespace FrameCPP::Version_8;

      FrameH*	f = dynamic_cast< FrameH* >( Frame.get( ) );
      if ( f )
      {
	FrameH::rawData_type* r = f->GetRawData( );
	if ( r )
	{
	  add_adc_channels( *r, Padding, ChannelCount);
	}
      }
    }
    break;
  default:
    break;
  }
}

static std::auto_ptr< frameh_type>
build_frame( version_type Version,
	     INT_4U ChannelCount )
{
  std::auto_ptr< frameh_type >	retval;

  retval = create_frameh( Version );

  const INT_4U cc_last = ChannelCount * 1024;
  int padding = 0;

  for ( INT_4U cc_cur = cc_last;
	cc_cur > 10;
	cc_cur /= 10 )
  {
    ++padding;
  }

  add_channels( Version, retval, padding, cc_last );
  return retval;
}

static std::auto_ptr< frameh_type>
create_frameh( version_type Version )
{
  std::auto_ptr< frameh_type >	retval;
  switch( Version )
  {
  case 8:
    retval = frameh< 8 >( );
    break;
  default:
    {
      std::ostringstream	msg;

      msg << "create_frameh does not support version "
	  << (int)Version
	  << " of the frame specification"
	;
      throw std::range_error( msg.str( ) );
    }
    break;
  }
  return retval;
}
